#ifndef _SPEINFO_H
#define _SPEINFO_H

// Datatype mapping

#define int_fp14 int
#define short_fp7 signed short
#define char_fp4 signed char

// Stub info from Spiral
#include "stub.h"

#define SIZE ROWS

#define SPUCHUNK ((SIZE/SPUS)/MULTIBUFFER_ITERATIONS)
#define SPUCHUNK_INBYTES (SPUCHUNK*DATATYPE_SIZEINBYTES)

#define MAX_DMA_SIZE 16384
#define DMACHUNK (MAX_DMA_SIZE/DATATYPE_SIZEINBYTES)
#define DMACHUNK_INBYTES MAX_DMA_SIZE

// Print debugging output, if DEBUG is defined
// #define DEBUG 1
// #define DEBUG1 1
// #define DEBUG_PPE

// Debugging output macro
#ifdef DEBUG
#include <stdio.h>
#include <assert.h>
#define debug(...)  { (void) fprintf (stderr, __VA_ARGS__); fflush(stderr); }
#else
#define debug(...)  ((void) 0)
#endif

#ifdef DEBUG_PPE
#include <stdio.h>
#include <assert.h>
#define debug_ppe(...)  { (void) fprintf (stderr, __VA_ARGS__); fflush(stderr); }
#else
#define debug_ppe(...)  ((void) 0)
#endif


#ifdef DEBUG1
#include <stdio.h>
#include <assert.h>
#define debug1(...)  { (void) fprintf (stderr, __VA_ARGS__); fflush(stderr); }
#else
#define debug1(...)  ((void) 0)
#endif

typedef struct {
  __attribute__ ((aligned(128))) unsigned long long size; //size of transform
  __attribute__ ((aligned(128))) DATATYPE_NO_QUOTES *X;
  __attribute__ ((aligned(128))) DATATYPE_NO_QUOTES *Y;
  __attribute__ ((aligned(128))) unsigned int cycle; //cycle count of the function
} data;

typedef struct {
  __attribute__ ((aligned(16))) unsigned long long X;
  __attribute__ ((aligned(16))) unsigned long long Y;
  __attribute__ ((aligned(16))) unsigned long long spu_ls[SPUS];
  __attribute__ ((aligned(16))) unsigned long long spu_ps[SPUS];
  __attribute__ ((aligned(16))) unsigned int spuid;
  __attribute__ ((aligned(16))) unsigned int num_runs; //number of times to run FUNC() in one shot
  __attribute__ ((aligned(16))) unsigned int cycles; //cycle count of the function
  __attribute__ ((aligned(16))) unsigned int timing_mode; //timing or verifying mode?

// ------------------------------------------------------------------------------
// ---------- This needs to be included dynamically. Okay, can be done.

// #include "twiddles-declare.h"
// ------------------------------------------------------------------------------


} spe_infostruct;

/* Conversion factor to obtain cycle count from timebase count.
 * Timebase for the IBM QS20 (3200/14MHz):
 * (See comment about TIMEBASE_MIN_PRECISION in ppu_timer.c)
 * IBM DALLAS machine
 */
// #define CYCLES_PER_TIMEBASE 223.4949
// #define TIMEBASE_MIN_PRECISION 50

// Timebase for the PS3 (3192MHz/79.8MHz):
#define CYCLES_PER_TIMEBASE 40.0
#define FREQ_IN_GHZ 3.192
#define TIMEBASE_MIN_PRECISION 200

// #define ALIGNED4(n)  (n%4)==0 ?  n : (n+( 4-(n% 4)))
// #define ALIGNED16(n) (n%16)==0 ? n : (n+(16-(n%16)))
// 
// #define SPU_DMA_SIZE(n) (n<=8) ? n : ALIGNED16(n)
// #define ALIGNED ALIGNED16

// PPU only
//FIXME: Move this into ppu/spu specific files or something. Looks ugly here.
#ifndef SPU
#define SSYNC __asm__ __volatile__  ("sync")
#define ISYNC __asm__ __volatile__  ("isync")

typedef struct {
  spe_context_ptr_t context;
  __attribute__ ((aligned(16))) spe_infostruct spe_info;
  unsigned int checksum;
  pthread_t sputhreadt;
} spu_t;

#endif
#endif // _SPEINFO_H
