#include <spu_mfcio.h>
#include <spu_intrinsics.h>

#define SPUID spe_info.spuid

// shared data defines between spu/ppu
#include "speinfo.h"

__attribute__ ((aligned ( 128))) spe_infostruct spe_info;

// TODO: Should alignment be 128 or 16? (can't see a performance diff)

static __attribute__ ((aligned (128))) DATATYPE_NO_QUOTES X[ SPUCHUNK ];

#ifdef OVERWRITEINPUTVEC
static DATATYPE_NO_QUOTES* Y = X;
#else
static __attribute__ ((aligned (128))) DATATYPE_NO_QUOTES Y[ SPUCHUNK ];
#endif

extern void INITFUNC(unsigned int spuid);
extern void FUNC(DATATYPE_NO_QUOTES *Y, DATATYPE_NO_QUOTES *X, unsigned int spuid);

unsigned int spuRecvSigValue = 0;
volatile int writeSignal[4] __attribute__ ((aligned (128)));
unsigned int spuPartner;
unsigned int check, total;
int  remainingChunk;


//---------------------- (old)
unsigned int tstart;
unsigned int tstop;
long cycles_storage[8];
int cs_count = 0;

void spu_applycounter(void) { cycles_storage[(cs_count)++] = spu_read_decrementer(); }
//----------------------

#define mask_signal 19
void sig_barrier(){

  debug("SPU[%u]: spu_ps[0] = %llu\n", spe_info.spuid, spe_info.spu_ps[0]);

  if(spe_info.spuid!=0){
    debug("SPU[%u]: Sending barrier signal.\n", spe_info.spuid);
    mfc_sndsig( ((char*) (&writeSignal) + 12), (spe_info.spu_ps[0]+12), mask_signal, 0, 0);
    debug("SPU[%u]: Waiting on barrier signal.\n", spe_info.spuid);
    spu_read_signal1(); // Will stall until read completes.
  }

  else{
    check = 1;
    debug("Master SPU 0: Waiting on signals.\n");
    while(check < total) {
      debug("SPU[%u]: Waiting on all signals. check=%u, total=%u\n", spe_info.spuid, check, total);
      check|=spu_read_signal1();
    }
    debug("Master SPU 0: All signals received.\n");

    //wake up every SPE
    for (int i=1; i<SPUS; i++) {
      debug("Master SPU 0: sending to %u\n", i);
      mfc_sndsig( ((char*) (&writeSignal) + 12), (spe_info.spu_ps[i]+12), mask_signal, 0, 0);
    }
  }
}

#include "spumacros.h"

void init_spu() {
  // Setup signal to write, and expected signals.
  writeSignal[3] = 1 << spe_info.spuid;
  total=(1<<SPUS)-1;
  spuRecvSigValue = (1 << spe_info.spuid) ^ ((1 << SPUS)-1);
  debug("SPU[%u]: init_spu(): SPUS=%d, total=%d\n", spe_info.spuid, SPUS, total);
}

//------------------------------------------------------------------------------
int main(unsigned long long spe_id, unsigned long long argp, unsigned long long envp) {
  unsigned int start, finish;

//  debug("SPU Begin. argp = %llu. sizeof(spe_infostruct)=%u\n", argp, sizeof(spe_infostruct));

  // Get spe_info contents from the PPU
  DMA_BL_GET_GENERIC(spe_info, argp, sizeof(spe_infostruct));
  debug("SPU[%u]: Begin. argp=%llx. argph=%x, argpl=%x. Num_runs=%u\n", spe_info.spuid, argp, mfc_ea2h(argp), mfc_ea2l(argp), spe_info.num_runs);

  init_spu();
  INITFUNC(SPUID);

  debug("SPU[%u]: init_spu() and INITFUNC() complete\n", spe_info.spuid);


  // Wait for message from PPU before proceeding.
  spu_read_in_mbox();

  debug("SPU[%u]: Mailbox message from PPU received. Proceeding to get X.\n", spe_info.spuid);

#ifndef VECTOR_IN_MEM
  debug("SPU[%u]: addr of X = %llu, size = %u\n",
      spe_info.spuid, (spe_info.X + (SPUCHUNK*spe_info.spuid)*DATATYPE_SIZEINBYTES), SIZE/SPUS);

  //*****************************************************************************
 
  // Get X: Max DMA size is 16k, so do multiple DMAs if needed:
  
  remainingChunk = SPUCHUNK_INBYTES;
  for (int i=0; i <= (SPUCHUNK/DMACHUNK); i++) {
    unsigned int dmaSize = remainingChunk < DMACHUNK_INBYTES ? remainingChunk : DMACHUNK_INBYTES;
    remainingChunk -= DMACHUNK_INBYTES; // Same as MAX_DMA_SIZE

    debug("SPU[%u]: X DMA_GET, DMACHUNK %u\n", SPUID, i);

    if (dmaSize != 0)
      DMA_ODD_BL_GET((X + (i*DMACHUNK)),
          (spe_info.X + (SPUCHUNK*spe_info.spuid*DATATYPE_SIZEINBYTES) + (i*DMACHUNK_INBYTES)), 
          dmaSize);
  }

  debug("SPU[%u]: X received via DMA. Proceeding with warmup.\n", spe_info.spuid);
#endif

  /* Warmup barrier. This is to try to get all the SPUs to start at
   * approximately the same time. */

  spu_writech(MFC_WrTagMask, -1);
  sig_barrier();

  if (spe_info.timing_mode == 1) {
    /* We're interested in timing. So we do a warm-up (can't do this for
     * verification, because we might end up clobbering X and/or Y.
     */
    debug("SPU[%u]: We're in timing mode. Beginning warmup run.\n", spe_info.spuid);
    FUNC(Y,X,SPUID);
    debug("SPU[%u]: Warmup complete. Proceeding with transform\n", spe_info.spuid);
  }

  spu_writech(MFC_WrTagMask, -1);
  sig_barrier();

  spu_write_decrementer(0x7fffffff);

  // Transform goes here. Reads X, Writes Y
  switch(spe_info.num_runs) {
    case 1:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    case 2:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    case 4:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    case 8:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    case 16:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    case 32:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    case 64:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);

      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    case 128:
      start = spu_read_decrementer();
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID); FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
      break;
    default:
      start = spu_read_decrementer();
      for (int i=0; i<spe_info.num_runs; i++)
        FUNC(Y,X,SPUID);
      finish = spu_read_decrementer();
  }

  debug("SPU[%u]: Transform complete.\n", spe_info.spuid);


#if (MULTIBUFFER_ITERATIONS <= 1)
#ifndef VECTOR_IN_MEM
  debug("SPU[%u]: DMA'ing Y back. addr of Y = %llu, size = %u\n",
      spe_info.spuid, (spe_info.Y + (SPUCHUNK*spe_info.spuid)*DATATYPE_SIZEINBYTES), SPUCHUNK);
  remainingChunk = SPUCHUNK_INBYTES;
  for (int i=0; i <= (SPUCHUNK/DMACHUNK); i++) {
    unsigned int dmaSize = remainingChunk < DMACHUNK_INBYTES ? remainingChunk : DMACHUNK_INBYTES;
    remainingChunk -= DMACHUNK_INBYTES; // Same as MAX_DMA_SIZE

    debug("SPU[%u]: Y DMA_PUT, DMACHUNK %u\n", SPUID, i);

    if (dmaSize != 0)
      DMA_ODD_BL_PUT((Y + (i*DMACHUNK)),
          (spe_info.Y + (SPUCHUNK*spe_info.spuid*DATATYPE_SIZEINBYTES) + (i*DMACHUNK_INBYTES)), 
          dmaSize);
  }

  debug("SPU[%u]: Y written via DMA.\n", spe_info.spuid);
#endif
#endif

  //*****************************************************************************

  debug("SPU[%u]: (start-finish) was %u\n", spe_info.spuid, (start-finish));

  spe_info.cycles = (start-finish);

  DMA_BL_PUT_GENERIC(spe_info, argp, sizeof(spe_infostruct));

  debug("SPU[%u]: DONE!\n", spe_info.spuid);
  return 0;
}

